#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#

from imblearn.under_sampling import RandomUnderSampler as imbRUS

from akangatu.abs.mixin.macro import asMacro
from akangatu.operator.unary.inop import In
from kururu.tool.enhancement.instance.sampling.abs.sampler import Sampler


class RUS_(Sampler):
    """Reduce the occurrence of all classes so as to be equal to the amount of instances in the minority class
    (default behavior)."""

    def __init__(self, strategy='not minority', replacement=False, seed=0):
        super().__init__(strategy=strategy, replacement=replacement, seed=seed)
        self.strategy = strategy
        self.replacement = replacement
        self.seed = seed

    def _algorithm_func(self):
        return lambda: imbRUS(sampling_strategy=self.strategy, replacement=self.replacement, random_state=self.seed)


class RUS(asMacro, RUS_):
    """Apply random under sampling to process inner data instead of using the outer data."""
    def _step_(self):
        return In(RUS_(**self.held))



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import re
from abc import ABC

import numpy as np
from aiuna.content.data import Data

from akangatu.distep import DIStep
from garoupa.util import flatten
from akangatu.transf.mixin.noop import asNoOp
from akangatu.transf.mixin.config import asUnitset


class AbsLog(asUnitset, DIStep, ABC):
    """Parent of Log and Report."""

    # TODO: Which matrix are prioritary by default?
    def __init__(self, text="Default log r=$r"):
        super().__init__(text=text)
        self.text = text

    @classmethod
    def _interpolate(cls, text, data):
        # TODO: Check how to set to re-prettify line breaks from numpy arrays.
        def samerow(M):
            return np.array_repr(M).replace("\n      ", "").replace("  ", "")

        def f(obj_match):
            field = obj_match.group(1)
            M = data.field(field, context=cls)
            if isinstance(M, np.float64):
                return str(M)
            try:
                if np.issubdtype(M, np.number):
                    return samerow(np.round(M, decimals=4))
            finally:
                return samerow(M)

        p = re.compile(r"\$([~a-zA-Z]+)")
        return cls._eval(p.sub(f, text), data)

    @classmethod
    def _eval(cls, text, data):
        txt = ""
        run = False
        expanded = [w.split("}") for w in ("_" + text + "_").split("{")]
        for seg in flatten(expanded):
            if run:
                code = "data." + seg
                try:
                    # Convert mapping through ~ operand.
                    if "~" in seg:
                        iterable, accessor = seg.split("~")
                        code = f"[item.{accessor} for item in data.{iterable}]"
                    if "^" in seg:
                        iterable, accessor = seg.split("^")
                        code = f"data.{iterable} ^ '{accessor}'"

                    # Data cannot be changed, so we don't use exec, which would accept dangerous assignments.
                    txt += str(eval(code))
                except Exception as e:
                    print(f"Problems parsing\n  {text}\nwith data\n  {data}\n" f"{data.history}\n!")
                    raise e
            else:
                txt += seg
            run = not run
        return txt[1:][:-1]


class Log(asNoOp, AbsLog):
    """Same as Report, but invisible in History"""

    def _process_(self, data: Data):  # overriding Report, to be invisible
        print(self._interpolate(self.text, data))
        return data



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import numpy as np
from sklearn.preprocessing import OneHotEncoder

from aiuna.content.data import Data
from aiuna.content.creation import nominal_idxs
from akangatu.distep import DIStep
from akangatu.transf.mixin.config import asConfigLess


class Binarize(asConfigLess, DIStep):  # TODO: other fields
    """Convert each nominal attribute to numeric ones by one-hot-encoding.

    All distinct values should be known before hand.
    There is no distinction between training and testing parts of the dataset.
    [this can be improved in the future by inheriting Transformer]

    For conversion from continuous attributes to binary attributes
    (same effect as nominal with two values) see step Discretize."""

    def _process_(self, data: Data):
        # REMINDER: Binarize will do nothing to numeric datasets, but the uuid still needs to be predictable.
        # So, the provided Data object should be "processed" anyway.
        def func():
            data_nominal_idxs = nominal_idxs(data.Xt)
            encoder = OneHotEncoder()
            if len(data_nominal_idxs) > 0:
                nom = encoder.fit_transform(data.X[:, data_nominal_idxs]).toarray()
                num = np.delete(data.X, data_nominal_idxs, axis=1).astype(float)
                return np.column_stack((nom, num))
            return data.X

        return data.update(self, X=func)


binarize = Binarize()



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from aiuna.content.data import Data
from akangatu.distep import DIStep
from akangatu.abs.mixin.macro import asMacro
from akangatu.fieldchecking import Forbid
from akangatu.operator.unary.inop import In
from kururu.tool.dataflow.autoins import AutoIns
from kururu.tool.evaluation.mixin.partitioning import withPartitioning


class Split1(DIStep, withPartitioning):
    """  """

    def __init__(self, i=0, stage="test", mode="holdout", splits=2, test_size=0.3, seed=0, fields="X,Y", _indices=None):
        config = locals().copy()
        del config["self"]
        del config["_indices"]
        self._indices = _indices
        self.i = i
        if stage == "train":
            self.istep = 0
        elif stage == "test":
            self.istep = 1
        else:
            print("Unknown stage:", stage)
            exit()

        withPartitioning.__init__(self, mode, config)
        DIStep.__init__(self, **config)

    def _process_(self, data: Data):
        def indices():
            if self._indices is None:
                self._indices = self.partitionings(data)[self.i][self.istep]
            return self._indices

        newmatrices = {}
        for f in self.fields:
            newmatrices[f] = (lambda f: lambda: data[f][indices()])(f)
        return data.update(self, **newmatrices)


class Split(withPartitioning, asMacro, DIStep):
    def __init__(self, i=0, mode="holdout", splits=2, test_size=0.3, seed=0, fields="X,Y", _indices=None):
        config = locals().copy()
        del config["self"]
        del config["_indices"]
        withPartitioning.__init__(self, mode, config)
        DIStep.__init__(self, **config)

    def _step_(self):
        # print("held:", self.held)
        internal = In(Split1(stage="train", **self.held))
        external = Split1(stage="test", **self.held)
        return Forbid("inner") * AutoIns * external * internal


split = Split()



from unittest import TestCase

from aiuna import Root
from aiuna.step.dataset import Dataset
from kururu import AutoIns
from kururu.tool.enhancement.instance.sampling.under.sample import Sample_, Sample


class TestSample(TestCase):
    def test__rus_(self):
        samplediris = Root >> (Dataset() * Sample_(n=120))
        self.assertEqual(120, len(samplediris.X))

    def test__rus(self):
        samplediris =  Root >> (Dataset() * Sample_(n=140))
        self.assertEqual(140, len((samplediris >> AutoIns * Sample).X))  # TODO: inhibit UserWarning
        self.assertEqual(100, len((samplediris >> AutoIns * Sample).inner.X))



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import numpy as np
from sklearn.metrics import accuracy_score

from aiuna.content.data import Data
from akangatu.abs.mixin.macro import asMacro
from akangatu.distep import DIStep
from akangatu.operator.unary.inop import In
from akangatu.transf.mixin.config import asUnitset
from kururu.tool.evaluation.mixin.functioninspection import withFunctionInspection


class Metrico(asUnitset, DIStep, withFunctionInspection):
    """Metric over fields.

    Only the-higher-the-better functions in Metric, but can be negated:
    '-accuracy' -> 'accuracy' * -1

    Developer: new metrics can be added just by following the pattern '_fun_xxxxx'
    where xxxxx is the name of the new metric.

    Parameters
    ----------
    functions
        Name of the function to use to evaluate data objects.
    target
        Name of the matrix with expected values.
    prediction
        Name of the matrix to be measured.
    """

    # noinspection PyDefaultArgument
    def __init__(self, functions=["accuracy"], target="Y",
                 prediction="Z"):  # TODO:change all default prameters to "mutable"
        super().__init__(functions=functions, target=target, prediction=prediction)
        self.functions = functions
        self.target, self.prediction = target, prediction
        self.selected = [self.function_from_name[name] for name in functions]

    def _process_(self, data: Data):
        newr = lambda: np.array([f(data, self.target, self.prediction) for f in self.selected])
        return data.update(self, r=newr)

    @staticmethod
    def _fun_accuracy(data, target, prediction):
        return accuracy_score(data[target], data[prediction])

    @staticmethod
    def _fun_history(data, target, prediction):
        return -len(list(data.history))


metric = Metrico


class Metricb(asMacro, Metrico):
    def _step_(self):
        metric = Metrico(**self.held)
        return metric * In(metric)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from abc import ABC, abstractmethod

from akangatu.transf.config import globalcache
from aiuna.content.data import Data
from akangatu.ddstep import DDStep


class Predictor(DDStep, ABC):
    """  """

    def _process_(self, data: Data):
        matrices = {"Z": lambda: self.model(data.inner).predict(data.X)}
        if "probability" in self.config and self.config["probability"]:
            matrices["P"] = lambda: self.model(data.inner).predict_proba(data.X)
        return data.update(self, **matrices)

    # @globalcache
    def model(self, data):
        """Predictor model induced on the training set.

        This cached method is needed because the so called "model" is called more than once inside process()."""
        model = self.algorithm
        model.fit(*data.Xy)
        return model

    @property
    def algorithm(self):
        return self._algorithm_func()()

    @abstractmethod
    def _algorithm_func(self):
        pass

    # x = 0.00001
    # l = []
    # for i in range(128):
    #     x = x * 1.1
    #     l.append(x)
    # print(l)

    """
default = SVM(tr).process(ts)
default = SVM(tr).process([tr, ts])
default = SVM(tr).process([ts1, ts2, ts3])
simples = SVM(tr, ...).process(d)
workf = File("iris") * Bin * Split(...) * dual(PCA(n=5) * dual(SVM(...))
expr = File("iris") * Split(...) * PCA(tr) * SVM(tr)

                                -> ts{tr}
expr = File("iris") * Split(...) * dual(NR(), Id()) * dual(PCA()) * dual(Id(), SVM())
Split(...) * PCA() * SVM()
= Chain(Split(...), dual(PCA())) * SVM()
= Chain(Split(...), dual(PCA()), dual(SVM()))

expr.process(ts)
expr.sample(ts).process(ts)
    
SVM(tr).process(ts)       ->  default
SVM(tr, ...).process(ts)  ->  configured

SVM(tr).sample()            ->  random
SVM(tr, ...).sample()       ->  partially configured/random
  
SVM()(tr).process(ts)     ->  default
SVM().sample()              ->  SVM(...)
SVM().sample()(tr)          ->  random
SVM(...)                    ->  lambda  
SVM(...)(tr).process(tr)  ->  configured
SVM(...)(tr).sample(ts)     ->  partially configured/random

Todo step transforma o data passado (externo).
Alguns dependem do data interno (trdata) como dado de entrada.
Para alterar dado interno, há dois modificadores: Inner e Both
Inner(NR()) -> aplica no trdata [mas transforma ambos! histórico: InnerNR muda o de fora e NR o de dentro]
Both(PCA()) -> aplica em ambos, mas treina sempre no interno [histórico: BothPCA muda externo e PCA muda interno] 
    """



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from dataclasses import dataclass
from typing import Iterator

from akangatu.linalghelper import islazy


@dataclass
class Result:
    value: float


class InterruptedStreamException(Exception):
    pass


class Accumulator(Iterator):
    """Cumulative iterator that returns a final/result value.

    The enclosed iterator should be finite."""

    def __init__(self, iterator, start, step_func, end_func, stream_exception=False):
        self.iterator = iterator
        self.start = start
        self.step_func = step_func
        self.end_func = end_func
        self.stream_exception = stream_exception

    def __next__(self):
        raise NotImplemented

    @property
    def result(self):
        if self.stream_exception:
            raise InterruptedStreamException
        try:
            return self._result
        except AttributeError as e:
            print("Stream not consumed!\nHINT: The result of summarizers are only accessible after Reduce.")
            exit()
            # raise e from None

    def __iter__(self):
        acc = self.start.copy()
        self.iterator = self.iterator() if islazy(self.iterator) else self.iterator
        try:
            for data in self.iterator:
                dic={"data":data}
                if not self.stream_exception:
                    dic = self.step_func(data, acc)
                    # if step is XXXXX:
                    #     self.stream_exception = True
                    # else:
                    if "inc" in dic:
                        # REMINDER: doesn't need to be thread-safe, since processing of iterator is always sequential
                        acc.append(dic["inc"])
                yield dic["data"]
        finally:
            if not self.stream_exception:
                self._result = self.end_func(acc)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#

from abc import ABC, abstractmethod

from sklearn.base import BaseEstimator

from akangatu.transf.config import globalcache

from akangatu.distep import DIStep


class Sampler(DIStep, ABC):
    """Base class for Step classes that have a resample method (currently applied only to X,y)."""
    def _process_(self, data):
        matrices = {"X": lambda: self.sampled(data)[0],
                    "Y": lambda: self.sampled(data)[1]}
        return data.update(self, **matrices)

    # @globalcache
    def sampled(self, data):
        """Sampled version of matrices.

        This cached method is needed because more than one matrix is calculated lazily inside process()."""
        if isinstance(self.algorithm, BaseEstimator):
            return self.algorithm.fit_resample(*data.Xy)
        else:
            return self.algorithm(*data.Xy)

    @property
    def algorithm(self):
        return self._algorithm_func()()

    @abstractmethod
    def _algorithm_func(self):
        pass



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from akangatu.distep import DIStep
# TODO: create Numpy step:  Numpy(method="sort", fromfield="A", tofield="B")
class Sort(DIStep):
    """Sort all provided matrices by a column(or row) of the first one."""

    def __init__(self, fields="U,X", byindex=0, reverse=False, along="rows"):
        super().__init__(fields=fields, byindex=byindex, reverse=reverse, along=along)
        self.fields = fields.split(",")
        self.along = along
        self.byindex = byindex
        self.reverse = reverse
        if along not in ["rows", "cols"]:
            print(self.name, f"Unknown type of entry to sort: by='{along}'. Should be 'rows' or 'cols'.")
            exit()  # TODO: generalize these checkings/errors

    def _process_(self, data):
        M = data.field(self.fields[0], context=self)
        if self.along == "rows":
            indices = M[:, self.byindex].argsort()
        else:
            indices = M[self.byindex, :].argsort()

        if self.reverse:
            indices = indices[::-1]

        dic = {}
        for field in self.fields:
            M = data.field(field, context=self)
            if self.along == "rows":
                dic[field] = M[indices, :]
            else:
                dic[field] = M[:, indices]

        return data.update(self, **dic)


class Sortr(Sort):
    """Shortcut for Sort(..., reverse=True)"""
    def __init__(self, fields="U,X", byindex=0, along="rows"):
        super().__init__(fields, byindex, True, along)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from numpy import mean, array

from akangatu import linalghelper
from akangatu.abs.mixin.macro import asMacro
from akangatu.distep import DIStep
from akangatu.fieldchecking import Forbid
from kururu.tool.evaluation.mixin.functioninspection import withFunctionInspection
from kururu.tool.manipulation.copy_ import Copy
from kururu.tool.stream.internal.accumulator import Accumulator
from akangatu.abs.mixin.fixedparam import asFixedParam


class Summ(asFixedParam, DIStep, withFunctionInspection):
    # Yes, we use "mutable" defaults because (it is immutable here and) better to show what to expect from the method.
    # ...and lists are more confortable to write/read than tuples.
    # noinspection PyDefaultArgument
    def __init__(self, stage="test", field="R", functions=["mean"]):
        super().__init__(stage=stage, field=field, functions=functions)
        self.functions = functions
        self.selected = [self.function_from_name[name] for name in functions]
        self.field = field
        if stage not in ["train", "test"]:
            print("Unknown stage:", stage)
            exit()
        self.stage = stage

    def _process_(self, data):
        if data.stream is None:
            print(f"{self.name} needs a Data object containing a stream.")
            print("Missing stream inside", data.id)
            exit()

        def step_func(data_, acc):
            if self.stage == "train":
                v = data_.inner.field(self.field, context=self)
            else:
                v = data_.field(self.field, context=self)
            return {"data": data_, "inc": linalghelper.mat2vec(v)}

        def end_func(acc):
            return [array(f(acc)) for f in self.selected]

        iterator = Accumulator(lambda: data.stream, start=[], step_func=step_func, end_func=end_func)
        return data.update(self, stream=lambda: iterator, S=lambda: iterator.result)

    @staticmethod
    def _fun_mean(values):
        return mean(values, axis=0)


class Summ2(asMacro, Summ):
    def _step_(self):
        train = Summ(stage="train", **self.held)
        test = Summ(stage="test", **self.held)
        return Forbid("inner") * train * Copy("S", "Si") * test

    # REMINDER: loop infinito no uuid ou json pode indicar presença de classe no config
